/********************************************************************
* FileName:         miwi_utilities.c
* Dependencies:     see include section
* Processor:        PIC18
*                   tested with 18F47J53, 18F27J53
* Complier:         Microchip XC8 v2.0 or higher
*
* Company:          SL Development
*
* Copyright and Disclaimer Notice for MyMaker Tools Software:
*
* Copyright  2022 MyMakerTools.  All rights reserved.
*
****************************************************************************
* File Description:
*
* This file provides the higher level MiWi functions above the standard MiWi
* framework.  These functions are based on multiple Microchip Demo programs.
*
* Change History:
*  Rev   Date         Author    Description
*  1.0   1-17-2022    JCW       Setup for MiWi app level functions
********************************************************************/

#include "system.h"             // standard system.h required by framework
#include "miwi_api.h"           // framework
#include "miwi_utilities.h"     // definitons for this file
#include "Messages.h"           // simple API for printing messsages to a console device
#include "Rogler.h"             // this file contains the ProcessMiWiQ function
#include "Timer5.h"             // contains basic TMR definitions and functionality


/************************ EXTERNAL VARIABLES **********************/
MIWI_DATA miwiData;

/**************************** Functions ***************************/
	
/********************************************************************
 * Function:        bool MiWiNetworkCommon(void)
 * PreCondition:    None
 * Input:           None
 * Output:          MiWi initialization status
 * Users:           Coordinators and end devices
 * Side Effects:    none
 * Overview:        This is common coordinator and end device code for initialization.
 * Note:            This function will init the Protocol chosen, set MyChannel
 *                  and then set the connection mode.
 *******************************************************************/
bool MiWi_NetworkCommon(void)                           
{
    bool result;

    result =  true;                                         // init the return variable
    /*******************************************************************/
    // Initialize Microchip proprietary protocol. Which protocol to use
    // depends on the configuration in ConfigApp.h
    /*******************************************************************/
    /*******************************************************************/
    // Function MiApp_ProtocolInit initialize the protocol stack. The
    // only input parameter indicates if previous network configuration
    // should be restored. In this simple example, we assume that the 
    // network starts from scratch.
    /*******************************************************************/
    MiApp_ProtocolInit(false);   
    if( MiApp_SetChannel(miwiData.myChannel) == false )
    {
        Message_ChannelError(miwiData.myChannel);           // send info to console
        result = false;                                     // connection failed
        return result;                                      // exit function with false
    }

    /*******************************************************************/
    // Function MiApp_ConnectionMode defines the connection mode. The
    // possible connection modes are:
    //  ENABLE_ALL_CONN:    Enable all kinds of connection
    //  ENABLE_PREV_CONN:   Only allow connection already exists in 
    //                      connection table
    //  ENABL_ACTIVE_SCAN_RSP:  Allow response to Active scan
    //  DISABLE_ALL_CONN:   Disable all connections. 
    /*******************************************************************/
    MiApp_ConnectionMode(ENABLE_ALL_CONN);
    Message_ChannelInfo(miwiData.myChannel, cConnecting);   // send info to console
    return result;                                          // return with true
}

/********************************************************************
 * Function:        bool CreateMiWiNetwork(void)
 * PreCondition:    None
 * Input:           None
 * Output:          MiWi connection status
 * Users:           coordinators only
 * Side Effects:    None
 * Overview:        This function will create a network
 * Note:            This will call start connection to create a network
 *******************************************************************/
MIWI_NET_STATES MiWi_CreateNetwork(void)
{
    bool result;
    MIWI_NET_STATES NetStatus;
    
    Printf("\r\nStarting MiWi Network...... \r\n");
    if (usbData.USBReady)                                   // if the USB connection is active, clear the print queue
        ProcessUSBOutput();
    result = MiWi_NetworkCommon();                          // execute the common network code
    if (result)                                             // if successful then continue
    {
        /*******************************************************************/
        // There is no network to join, we need to start a new 
        // network by calling function MiApp_StartConnection
        //
        // The first parameter is the mode of start connection. There are 
        // two valid connection modes:
        //   - START_CONN_DIRECT        start the connection on current 
        //                              channel
        //   - START_CONN_ENERGY_SCN    perform an energy scan first, 
        //                              before starting the connection on 
        //                              the channel with least noise
        //   - START_CONN_CS_SCN        perform a carrier sense scan 
        //                              first, before starting the 
        //                              connection on the channel with 
        //                              least carrier sense noise. Not
        //                              supported for current radios
        //
        // The second parameter is the scan duration, which has the same 
        //     definition in Energy Scan. 10 is roughly 1 second. 9 is a 
        //     half second and 11 is 2 seconds. Maximum scan duration is 
        //     14, or roughly 16 seconds.
        //
        // The third parameter is the channel map. Bit 0 of the 
        //     double word parameter represents channel 0. For the 2.4GHz 
        //     frequency band, all possible channels are channel 11 to 
        //     channel 26. As the result, the bit map is 0x07FFF800. Stack 
        //     will filter out all invalid channels, so the application 
        //     only needs to pay attention to the channels that are not 
        //     preferred.
        /*******************************************************************/
        MiApp_ProtocolInit(false);   
        MiApp_StartConnection(START_CONN_DIRECT,0, 0);
        Printf("MiWi Connection Started\r\n");

        result = (ConnectionTable[0].status.bits.isValid);      // check for connections
        if (result)                                             // if connections
        {
            NetStatus = MIWI_NET_STATE_FORMING;
        }
        else
        {
            NetStatus = MIWI_NET_STATE_NOCONNECT;
        }
    }
    else                                                        // the common code failed
    {
        NetStatus = MIWI_NET_STATE_FAILED;
    }
    if (usbData.USBReady)                                       // if the USB connection is active, clear the print queue
        ProcessUSBOutput();
    return NetStatus;    
}
/********************************************************************
 * Function:        bool JoinMiWiNetwork(void)
 * PreCondition:    None
 * Input:           None
 * Output:          MiWi connection status
 * Users:           end devices only
 * Side Effects:    if return is false, TMR3 is restarted
 * Overview:        This function attempts to find a network to connect to
 * Note:            This should never be needed, but here is just in case.
 *                  This function joins an existing network and the Master
 *                  always creates the network, at least for now.
 *******************************************************************/
bool MiWi_JoinNetwork(void)
{
    bool result;
    uint8_t ConnectStatus;
    
    Printf("\r\nLooking for MiWi Network...... \r\n");
    if (usbData.USBReady)                                       // if the USB connection is active, clear the print queue
        ProcessUSBOutput();
    result = MiWi_NetworkCommon();                              // execute the common network code
    
    /*******************************************************************/
    // Function MiApp_EstablishConnection try to establish a new 
    // connection with peer device. 
    // The first parameter is the index to the active scan result, which 
    //      is acquired by discovery process (active scan). If the value 
    //      of the index is 0xFF, try to establish a connection with any 
    //      peer.
    // The second parameter is the mode to establish connection, either 
    //      direct or indirect. Direct mode means connection within the 
    //      radio range; Indirect mode means connection may or may not 
    //      in the radio range. 
    // Return value is an index to the connection table.  0xFF means that
    //      no connection was made
    /*******************************************************************/

    ConnectStatus = MiApp_EstablishConnection(0xFF, CONN_MODE_DIRECT);
    if (ConnectStatus != 0xFF)                                  // there was a connection
    {	
        result = true;
        Message_ChannelInfo(miwiData.myChannel, 1);             // send info to the console
        /*******************************************************************/
        // Function DumpConnection is used to print out the content of the
        // Connection Entry on the console. It may be useful in 
        // the debugging phase.
        // The only parameter of this function is the index of the  
        // Connection Entry. The value of 0xFF means to print out all
        // valid Connection Entry; otherwise, the Connection Entry
        // of the input index will be printed out.
        /*******************************************************************/
        DumpConnection(0xFF);
        Printf("Ready to connect!\r\n");
    }
    else
    {
        result = false;
        Printf("No connection found!\r\n");
    }
    if (usbData.USBReady)                                       // if the USB connection is active, clear the print queue
        ProcessUSBOutput();
    return result;
}

/*********************************************************************
* Function:         void MiWiChannelScan(void)
* PreCondition:     init MiWi Network
* Input:            myChannel
* Output:           channel status
* Users:            coordinators and end devices
* Side Effects:	    none
* Overview:         this prints out the channel status
* Note:			    none
********************************************************************/
void MiWi_ChannelScan(uint8_t myMiWiChannel)
{
    //bool Tx_Success = false;
    uint8_t channel, scanresult;
    uint32_t ChannelMask;
        
    if (MiApp_SetChannel(myMiWiChannel) == false)               // if the set channel returned false
    {
        Message_ChannelError(myMiWiChannel);                    // send error info to console
        return;                                                 // we are done then return
    }
    //Scan for networks (Testing TX and RX functionality        // channel set worked, so scan all channels
    ChannelMask = Mask_Ch11;                                    // set channel mask to the lowest channel
    for (channel=11;channel<27;channel++)                                     
    {
        scanresult = MiApp_SearchConnection(10, ChannelMask);   // check the channel and store in scanresult
        ChannelMask = ChannelMask << 0x01;                      // move to next channel
        Message_ScanResults(channel,scanresult);                      // send results to console
        while (usbData.USBQCount > 0)                           // if the USB connection is active, clear the print queue
            ProcessUSBOutput();
    }
}

/*********************************************************************
* Function:         uint8_t CountConnections(void)
* PreCondition:     init MiWi Network
* Input:            none
* Output:           connection count
* Users:            coordinators only should use this
* Side Effects:	    none
* Overview:         this counts the connections
* Note:			    none
********************************************************************/
uint8_t MiWi_CountConnections(void)
{
	uint8_t result = 0;                                         // init the connection count to zero
    uint8_t index;
	
    for (index=0; index < CONNECTION_SIZE; index++)             // check each location in the connection table
	{
		if (ConnectionTable[index].PeerInfo[0] > 0)             // check if there is a connection here
		{
			result++;                                           // increment the connection count
        }
    }
    return result;                                              // return the connection count
}

/*********************************************************************
* Function:         void MiWiTasks(void)
* PreCondition:     uses TMR5 
* Input:            none
* Output:           none
* Side Effects:	    none
* Overview:         performs the required MiWi tasks in a case statement
* Note:			    none
********************************************************************/
void MiWi_AppTasks(void)
{
	uint8_t result = 0;
    uint8_t index;
	
    switch (miwiData.appState) 
    {
        case MIWI_STATE_INIT:                           // initial state, always start here for both coordinators and end devices
        {
            miwiData.MiWiConnection = false;                // initialize the variables 
            miwiData.NetworkFormed = false;
            miwiData.myChannel = cMyChannel;
            miwiData.netState = MIWI_NET_STATE_INIT;
            miwiData.netState_OLD = miwiData.netState;
            #if (RoglerBoardType < btZenaMaster)            // this is for the end device 
                TMR5Interval = TMR5_10sec;                  //set the timer var to 10 sec                
                Timer5IntDelay(TMR5Interval);               //start timer interval @ 10 sec, this is a non-blocking call
                miwiData.appState = MIWI_STATE_WAIT_2_JOIN; // end device goes here next
            #else                                           // this is for the coordinator
                miwiData.appState = MIWI_STATE_CREATE;      // coordinator device goes here next
            #endif
            break;    
        }
        case MIWI_STATE_CREATE:                         // coordinators only
        {
            miwiData.MiWiConnection = false;            // init the var
            miwiData.netState = MiWi_CreateNetwork();   // perform the network creation
            switch (miwiData.netState)
            {
                case MIWI_NET_STATE_FAILED:
                {                                       // stay in create state
                    miwiData.netState = MIWI_NET_STATE_FAILED;
                    break;
                }
                case MIWI_NET_STATE_NOCONNECT:          // there is no connection
                {
                    miwiData.appState = MIWI_STATE_RETRY_CREATE;// goto retry create state
                    miwiData.netState = MIWI_NET_STATE_INIT;
                    break;
                }
                case MIWI_NET_STATE_FORMING:            // someone has connected
                {
                    miwiData.MiWiConnection = true;
                    miwiData.appState = MIWI_STATE_WAIT_4_CONNECTS;// go to this state
                    miwiData.netState = MIWI_NET_STATE_FORMING;
                    break;
                }
            }
            break;
        }
        case MIWI_STATE_RETRY_CREATE:                   // coordinators only
        {
            if( MiApp_MessageAvailable() )                  // check for messages
                MiApp_DiscardMessage();                     // and clear any messages
            miwiData.MiWiConnection = (ConnectionTable[0].status.bits.isValid);  //has anyone connected
            if (miwiData.MiWiConnection)                    // if somebody connected
            {    
                miwiData.appState = MIWI_STATE_WAIT_4_CONNECTS;// goto this state and wait for expected end devices to connect
                miwiData.netState = MIWI_NET_STATE_FORMING;
            }
            break;
        }
        case MIWI_STATE_WAIT_4_CONNECTS:                // coordinators only
        {
            result = MiWi_CountConnections();               // how many connections are there
            if (result > miwiData.MiWiConnectionCount)      // if this number is greater than last check
                DumpConnection(0xFF);                       // then dump the connection to the console
            miwiData.MiWiConnectionCount = result;          // update the connection count
            if (miwiData.MiWiConnectionCount >= miwiData.ExpectedCount) // if connection count is greater than the expected count
            {
                miwiData.NetworkFormed = true;              // the network has formed
                miwiData.appState = MIWI_STATE_ACTIVE;      // goto the active state
                miwiData.netState = MIWI_NET_STATE_FORMED;
            }
            else                                            // if the connection count is not greater than the expected count
                miwiData.appState = MIWI_STATE_RETRY_CREATE;// then go back to retry create and keep checking
            break;
        }
        case MIWI_STATE_WAIT_2_JOIN:                    // end devices only
        {
            if (TMR5Fired)                                  //wait for 10 sec before joining attempt, let coordinator startup
            {
                miwiData.appState = MIWI_STATE_JOIN;        // now that 10 sec have expired, attempt to join a network
                TMR5Fired = false;                          // reset the variable
                TMR5Interval = TMR5_1sec;                   // return the TMR5 interval to 1 sec                
            }
            break;
        }
        case MIWI_STATE_JOIN:                           // end devices only
        {
            miwiData.MiWiConnection = MiWi_JoinNetwork();   // perform the network joining
            if (miwiData.MiWiConnection)                    // if there was a network connection
                miwiData.appState = MIWI_STATE_ACTIVE;      // go to the active state
            else                                            // if there was no network connection
                miwiData.appState = MIWI_STATE_RETRY_JOIN;  // then go to the rejoin state
            break;
        }
        case MIWI_STATE_RETRY_JOIN:                     // end devices only
        {                                                   // this is a place holder if more elaborate re-joining is needed
            miwiData.appState = MIWI_STATE_JOIN;            // go back to the join state       
            break;
        }
        case MIWI_STATE_ACTIVE:                             // coordinators and end devices
        {
            /*******************************************************************/
            // Function MiApp_MessageAvailable returns a boolean to indicate if 
            // a packet has been received by the transceiver. If a packet has
            // been received, all information will be stored in the rxFrame, 
            // structure of RECEIVED_MESSAGE.
            /*******************************************************************/
            if( MiApp_MessageAvailable() )
            {
                /*******************************************************************/
                // If a packet has been received, handle the information available 
                // in rxMessage.
                /*******************************************************************/
                ProcessMiWiQ();                         // this function will process the message as required
                /*******************************************************************/
                // Function MiApp_DiscardMessage is used to release the current 
                // received packet.  After calling this function, the stack can 
                // start to process the next received frame 
                /*******************************************************************/  
                MiApp_DiscardMessage();
            }
            break;
        }
        case MIWI_STATE_IDLE:                           // coordinators and end devices
        {                                                   // placeholder
            break;
        }
            
    }
    if (miwiData.netState != miwiData.netState_OLD)         // if the net state has changed, update the status
    {
        UpdateMasterStatus();                               // update the coordinator status to the console
        miwiData.netState_OLD = miwiData.netState;          // update the net state
    }
}
